from abc import ABC, abstractmethod
from typing import Iterable
from ...document import Document

class BaseEmbeddingModel(ABC):
    
    @property
    @abstractmethod
    def name(self) -> str:
        """Model name.
        """
        pass

    @property
    @abstractmethod
    def vec_dim(self) -> int:
        """Dimension of the embedding vector space.
        """
        pass

    @abstractmethod
    def embed_text(self, text: str) -> list[float]:
        pass

    @abstractmethod
    def embed_texts(self, texts: Iterable[Document]) -> list[list[float]]:
        pass
    
    

    


from .doc2text import Document2TextTransformer
from .sentence_transformer import SentenceTransformerEmbeddingModel

__all__ = [
    'Document2TextTransformer',
    'SentenceTransformerEmbeddingModel'
]


from sentence_transformers import SentenceTransformer
from typing import Self, Optional, Iterable
from .base import BaseEmbeddingModel

class SentenceTransformerEmbeddingModel(BaseEmbeddingModel, SentenceTransformer):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._name = None
    
    @property
    def name(self) -> str:
        name = self._name
        assert isinstance(name, str)
        return name
    
    @property
    def vec_dim(self) -> int:
        vec_dim = self.get_sentence_embedding_dimension()
        assert isinstance(vec_dim, int)
        return vec_dim
    
    @classmethod
    def from_model_name(cls, model_name: str = 'all-mpnet-base-v2') -> Self:
        model = cls(model_name_or_path=model_name)
        model._name = model_name
        return model
    
    def embed_text(self, text: str) -> Optional[list[float]]:
        
        # compute embedding of the given document
        vec = self.encode(text)
        
        # convert to a list
        vec = vec.tolist()
        
        return vec
    
    def embed_texts(self, texts: Iterable[str]) -> list[list[float]]:
        
        # compute embeddings of all documents
        vecs = self.encode(texts)
        
        # covert to a list of vectors
        vecs = vecs.tolist()
        
        return vecs



from typing import Any, Self, Optional
from .id import DocumentID

class Document(dict):
    
    KEY_OF_TEXT = 'text'
    
    def __init__(self, *args, **kwargs) -> None:
        
        super().__init__(*args, **kwargs)
        
        self._id = None
        self._text = None
    
    @property
    def id(self) -> Optional[DocumentID]:
        return self._id
    
    @id.setter
    def id(self, new_id: DocumentID) -> None:
        
        assert isinstance(new_id, DocumentID)
        self._id = new_id
    
    @property
    def text(self) -> Optional[str]:
        
        return self.get_text()
    
    @classmethod
    def from_dict(cls, document: dict) -> Self:
        document = cls(document)
        return document
    
    def get_text(self) -> Optional[str]:
        return self.get(self.KEY_OF_TEXT, None)
            
    def to_dict(self, with_id: bool = False) -> dict:
        
        document_dict = self.copy()
        
        if with_id:
            if self.id is None:
                document_dict['id'] = None
            else:
                document_dict['id'] = self.id.value
            
        return document_dict
    
    


from typing import Optional, Iterable
from itertools import filterfalse
from .document import Document, DocumentID
from .docdb.base import BaseDocumentDatabaseClient
from .vecdb.base import BaseVectorDatabaseClient

class KnowledgeBaseClient:
    
    def __init__(
            self,
            doc_db_client: BaseDocumentDatabaseClient,
            vec_db_client: BaseVectorDatabaseClient
        ) -> None:
        
        self._doc_db_client = doc_db_client
        self._vec_db_client = vec_db_client
    
    @property
    def doc_db_client(self) -> BaseDocumentDatabaseClient:
        return self._doc_db_client
    
    @property
    def vec_db_client(self) -> BaseVectorDatabaseClient:
        return self._vec_db_client
    
    @property
    def doc_db_collection_name(self) -> str:
        return self._doc_db_client.collection_name
    
    @property
    def vec_db_collection_name(self) -> str:
        return self._vec_db_client.collection_name
    
    def find_document(self, *args, **kwargs) -> Optional[Document]:
        
        return self._doc_db_client.find_document(*args, **kwargs)
    
    def find_documents(self, *args, **kwargs) -> list[Document]:
        
        return self._doc_db_client.find_documents(*args, **kwargs)
    
    def insert_document(
            self, 
            document: Document,
            with_embedding: bool = False
        ):
        
        # do nothing if the document has an ID,
        # which means it already exits in the document database
        if document.id is not None:
            return
        
        # insert into document database
        # and then get the inserted ID
        document_id = self._doc_db_client.insert_document(document)
        
        # store the vector embedding of the document if required
        if with_embedding:
            
            # set the ID
            document.id = document_id
            
            # insert into vector database
            self._vec_db_client.upsert_document(document)
    
    def update_document_embeddings(self, documents: Iterable[Document]) -> list[DocumentID]:
        
        document_ids = self._vec_db_client.upsert_documents(documents)
        
        return document_ids
    
    def insert_documents(
            self, 
            documents: Iterable[Document],
            with_embedding: bool = False
        ) -> dict[str, list[DocumentID]]:
        
        # filter out those documents that already have IDs
        documents = list(filter(
            lambda document: document.id is None,
            documents
        ))
        
        # insert multiple documents into document database
        # and get the inserted IDs
        inserted_document_ids_in_doc_db = self._doc_db_client.insert_documents(documents)
        
        # set IDs
        for i, _ in enumerate(documents):
            documents[i].id = inserted_document_ids_in_doc_db[i]
        
        # store all vector embeddings of the document if required
        if with_embedding:
            inserted_document_ids_in_vec_db = self._vec_db_client.upsert_documents(documents)
            return {
                'doc_db': inserted_document_ids_in_doc_db,
                'vec_db': inserted_document_ids_in_vec_db
            }
        
        return {
            'doc_db': inserted_document_ids_in_doc_db,
            'vec_db': []
        }
        
    def update_documents(
            self, 
            documents: Iterable[Document],
            do_insert: bool = True
        ) -> dict[str, list[DocumentID]]:
        
        update_result = {
            'doc_db': [],
            'vec_db': []
        }
        
        # update documents in document database
        document_ids = self._doc_db_client.update_documents(
            documents=documents, 
            do_insert=do_insert
        )
        update_result['doc_db'] = document_ids
        
        # find the corresponding documents
        documents = self._doc_db_client.find_documents_by_ids(document_ids)
        
        # update embedding vectors
        self.update_document_embeddings(documents)
        update_result['vec_db'] = document_ids
        
        return update_result
        
    def delete_documents(
            self, 
            documents: Iterable[Document]
        ) -> dict[str, list[DocumentID]]:
        
        deletion_result = {
            'doc_db': [],
            'vec_db': []
        }
        
        # delete from document database
        deleted_document_ids = self._doc_db_client.delete_documents(documents)
        deletion_result['doc_db'] = deleted_document_ids
        
        # delete from vector database
        deleted_document_ids = self._vec_db_client.delete_documents_by_ids(deleted_document_ids)
        deletion_result['vec_db'] = deleted_document_ids
        
        return deletion_result

    def retrieve_similar_documents(
            self, 
            query: str,
            n_similar_documents: int
        ) -> list[Document]:
        
        # retrieve similar document IDs from vector database
        similar_document_ids = self._vec_db_client.retrieve_similar_document_ids(
            query,
            n_similar_documents
        )
        
        # find the actual documents in the document database
        similar_documents = self._doc_db_client.find_documents_by_ids(similar_document_ids)
        
        return similar_documents
    


from sklearn.base import BaseEstimator, TransformerMixin
from typing import Self, Optional, Iterable
from ...document import Document, DocumentID

class Document2TextTransformer(BaseEstimator, TransformerMixin):
    
    def __init__(self) -> None:
        super().__init__()
        
        self._document_ids = None
    
    @property
    def document_ids(self) -> Optional[list[DocumentID]]:
        return self._document_ids
    
    def fit(self, document: Document) -> Self:
        return self 
    
    def transform(
            self, 
            document_or_documents: Document | Iterable[Document]
        ) -> Optional[str] | list[str]:
        """Transform each document its text content.

        Parameters
        ----------
            document_or_documents (Document | Iterable[Document]): A single document
            or multiple documents.

        Returns
        -------
            Optional[str] | list[str]: 
            - If a single document is passed in, 
            then a string of its text content will be returned.
            - If multiple documents are passed in, 
            then a list of text contents will be returned.
        """
        
        if isinstance(document_or_documents, Document):
            document = document_or_documents
            text = self.get_text(document)
            self._document_ids = None
            return text
        
        documents = document_or_documents
        texts: list[str] = []
        self._document_ids: list[DocumentID] = []
        for document in documents:
            text =  self.get_text(document)
            if text is not None:
                texts.append(text)
                self._document_ids.append(document.id)
        
        return texts

    def get_text(self, document: Document) -> Optional[str]:
        
        assert isinstance(document, Document)
        return document.get('text', None)
    


from typing import (
    Any, 
    Self,
    Optional,
    Iterable
)
from itertools import filterfalse
from natia.kb.document import DocumentID
import pymongo
from pymongo.database import Database
from pymongo.collection import Collection
from .base import BaseDocumentDatabaseClient
from ..document import Document, DocumentID

class MongoDBClient(BaseDocumentDatabaseClient, pymongo.MongoClient):
    
    def __init__(self, *args, **kwargs) -> None:
        
        super().__init__(*args, **kwargs)
        
        self._database: Optional[Database] = None
        self._collection: Optional[Collection] = None
    
    @property
    def collection_name(self) -> str:
        return self._collection.name
    
    @classmethod
    def connect_to_host_port(cls, host: str, port: int) -> Self:
        """Connect to MongoDB host on a given port.

        Parameters
        ----------
            host (str): Host.
            port (int): Port.

        Returns
        -------
            Self: MongoDB client.
        """
        
        client = cls(
            host=host,
            port=port
        )
        
        return client
    
    def open_database(
            self,
            database_name: str
        ) -> Self:
        
        self._database = self.get_database(database_name)
        
        return self
    
    def open_collection(
            self,
            collection_name: str
        ) -> Self:
        
        if self._database is None:
            raise RuntimeError('no database is opened')
        
        self._collection = self._database.get_collection(collection_name)
        
        return self
        
    def insert_documents(self, documents: Iterable[Document]) -> list[DocumentID]:
        
        # filter out those documents that have IDs
        # since they are already in the collection
        documents = self.filter_out_documents_with_ids(documents)
        
        # stop here if there is no document to insert
        if len(documents) == 0:
            return []
        
        # get insertion result from MongoDB
        insertion_result = self._collection.insert_many(documents)
        
        # convert to document IDs
        inserted_ids = list(map(
            DocumentID,
            insertion_result.inserted_ids
        ))
        
        return inserted_ids
        
    def find_document(self, filter: Optional[Any] = None) -> Optional[Document]:
        
        # find document in the collection
        document: Optional[dict] = self._collection.find_one(filter)
        
        # no document is found
        if document is None: 
            return None

        assert isinstance(document, dict)
        
        document_id = DocumentID(document.pop('_id'))
        
        # convert to document instance
        document: Document = Document.from_dict(document)
        
        # set document ID
        document.id = document_id
        
        return document
    
    def find_documents(self, filter: Optional[Any] = None) -> list[Document]:
        
        # find documents in the collection
        fetched_documents = self._collection.find(filter)
        
        # a list of Document instances to return
        documents = []
        
        for document in fetched_documents:
            
            document_id = DocumentID(document.pop('_id'))
        
            # convert to document instance
            document: Document = Document.from_dict(document)
        
            # set document ID
            document.id = document_id
            
            documents.append(document)
        
        return documents
    
    def find_document_by_id(self, id: DocumentID) -> Optional[Document]:
        
        return self.find_document(id.to_objectid())
    
    def find_documents_by_ids(self, ids: Iterable[DocumentID]) -> list[Document]:
        
        # use find_many method provided by MongoDB 
        # may not be helpful since the number of documents to find 
        # is not large,
        # and most importantly, 
        # the returned documents must be in the same order
        # as the input IDs
        
        documents: list[Document] = []
        for document_id in ids:
            document = self.find_document_by_id(document_id)
            if document is not None:
                documents.append(document)
        
        return documents
    
    def update_documents(
            self, 
            documents: Iterable[Document],
            do_insert: bool = True
        ) -> list[DocumentID]:
        
        document_ids: list[DocumentID] = []
        
        for document in documents:
            
            # the document is not in the collection yet
            if document.id is None:
                if do_insert:
                    document_id = self.insert_document(document)
                    if document_id is not None:
                        document_ids.append(document_id)
                else:
                    continue
            
            # update the document
            self._collection.update_one(
                filter={
                    '_id': document.id.to_objectid()
                },
                update={
                    '$set': document
                }
            )
            
            document_ids.append(document.id)
        
        return document_ids
    
    def delete_documents(
            self, 
            documents: Iterable[Document]
        ) -> list[DocumentID]:
        
        # extract docuemnt IDs
        document_ids = list(map(
            lambda document: document.id, 
            documents
        ))
        
        # get deleted document IDs
        deleted_document_ids = self.delete_documents_by_ids(document_ids)
        
        return deleted_document_ids
        
    def delete_documents_by_ids(
            self, 
            ids: Iterable[DocumentID]
        ) -> list[DocumentID]:
        
        # get nonempty IDs
        ids = self.filter_out_empty_document_ids(ids)
        
        # convert each ID to ObjectId instance
        document_ids_as_objectids = list(map(
            lambda id: id.to_objectid(),
            ids
        ))
        
        # delete documents
        self._collection.delete_many({
            '_id': {
                '$in': document_ids_as_objectids
            }
        })
        
        return ids
    
    def drop_collection(self, collection_name: str) -> None:
        
        if self._database is None:
            raise RuntimeError('no database is opened')
        
        self._database.drop_collection(collection_name)
      
        
    



import unittest
from pprint import pprint
from natia.kb import KnowledgeBaseClient
from natia.kb.document import Document, DocumentID
from natia.kb.docdb import MongoDBClient
from natia.kb.vecdb import QdrantClient
from natia.kb.vecdb.doc2vec import (
    Document2TextTransformer,
    SentenceTransformerEmbeddingModel
)

class KnowledgeBaseTester(unittest.TestCase):
    
    def test_basics(self):
        
        print()
        
        # set up document database
        doc_db_client = MongoDBClient.connect_to_host_port(
            host='localhost',
            port=27017
        ).open_database(
            database_name='test-document-database'
        ).open_collection(
            collection_name='test-document-collection'
        )
        
        # set up vector database
        
        document2text_transformer = Document2TextTransformer()
        
        embedding_model = SentenceTransformerEmbeddingModel.from_model_name('all-MiniLM-L6-v2')
        
        vec_db_client = QdrantClient.connect_to_host_port(
            host='localhost',
            port=6333
        ).open_collection(
            collection_name='test-vector-database-collection',
            document2text_transformer=document2text_transformer,
            embedding_model=embedding_model
        )
        
        # set up knowledge base
        knowledge_base_client = KnowledgeBaseClient(
            doc_db_client=doc_db_client,
            vec_db_client=vec_db_client
        )
        
        result = knowledge_base_client.insert_documents(
            documents=[],
            with_embedding=True
        )
        print('Insertion result:')
        pprint(result)
        print()
        
        result = knowledge_base_client.insert_documents(
            documents=[
                Document(user='Isaac', age=23, text='Hello, MongoDB!'),
                Document(user='John', age=50, text='I will serve. I will be of service')
            ]
        )
        print('Insertion result:')
        pprint(result)
        print()
        
        result = knowledge_base_client.insert_documents(
            documents=[
                Document(user='Isaac', text='Hello, Natia!'),
                Document(user='Albert', age=89, text='I taught mathematics.'),
                Document(user='Tom', text='I invented the language TOML.'),
                Document(user='Jex', gpa=3.1, text='I am a front-end engineer?'),  
            ],
            with_embedding=True
        )
        print('Insertion result:')
        pprint(result)
        print()
        
        # delete documents
        documents_to_delete = knowledge_base_client.find_documents({
            'text': {
                '$regex': 'Hello'
            }
        })
        print('Documents to delete:')
        pprint(documents_to_delete)
        result = knowledge_base_client.delete_documents(documents_to_delete)
        print('Deletion result:')
        pprint(result)
        print()
        
        # retrieve similar documents
        documents = knowledge_base_client.retrieve_similar_documents(
            query='algebra',
            n_similar_documents=2
        )
        print('Similar documents:')
        pprint(documents)
        print()
        
        # udpate documents
        documents[0]['value'] = 100
        update_result = knowledge_base_client.update_documents(documents)
        print('Update result:')
        pprint(update_result)
        pprint(knowledge_base_client._doc_db_client.find_documents_by_ids(update_result['doc_db']))
        print()
        
        updated_document_ids = knowledge_base_client.update_document_embeddings(documents)
        pprint(updated_document_ids)
        
        # clean up
        
        doc_db_client.drop_collection(
            collection_name='test-document-collection'
        )
        
        vec_db_client.drop_collection(
            collection_name='test-vector-database-collection'
        )

if __name__ == '__main__':
    
    unittest.main()
    


from abc import ABC, abstractmethod
from typing import Optional, Iterable
from itertools import filterfalse
from ..document import Document, DocumentID

class BaseDocumentDatabaseClient(ABC):
    
    @property
    @abstractmethod
    def collection_name(self) -> str:
        pass
    
    @staticmethod
    def filter_out_documents_with_ids(
            documents: Iterable[Document]
        ) -> list[Document]:
        
        documents = list(filter(
            lambda document: document.id is None,
            documents
        ))
        
        return documents
    
    @staticmethod
    def filter_out_documents_without_ids(
            documents: Iterable[Document]
        ) -> list[Document]:
        
        documents = list(filterfalse(
            lambda document: document.id is None,
            documents
        ))
        
        return documents
    
    @staticmethod
    def filter_out_empty_document_ids(
            ids: Iterable[DocumentID]
        ) -> list[DocumentID]:
        
        ids = list(filterfalse(
            lambda id: id is None,
            ids
        ))
        
        return ids
    
    @abstractmethod
    def insert_documents(self, documents: Iterable[Document]) -> list[DocumentID]:
        pass
    
    def insert_document(self, document: Document) -> Optional[DocumentID]:
        
        document_ids = self.insert_documents([document])
        if len(document_ids) > 0:
            return document_ids[0]
        else:
            return None
    
    @abstractmethod
    def find_document(self, *args, **kwargs) -> Optional[Document]:
        """Find one document in the database satisfying some requirements.

        Returns
        -------
            Optional[Document]: None if no ducument is found.
        """
        pass
    
    @abstractmethod
    def find_documents(self, *args, **kwargs) -> list[Document]:
        """Find all documents in the databae satisfying some requirements.

        Returns
        -------
            list[Document]: An empty list if no ducument is found.
        """
        pass
    
    @abstractmethod
    def find_document_by_id(self, id: DocumentID) -> Optional[Document]:
        """Find one document in the databae by its ID.

        Parameters
        ----------
            id (DocumentID): Document ID.

        Returns
        -------
            Optional[Document]: None if no ducument is found.
        """
        pass
    
    def find_documents_by_ids(self, ids: Iterable[DocumentID]) -> list[Document]:
        """Find one document in the databae by its ID.

        Parameters
        ----------
            ids (Iterable[DocumentID]): Document IDs.
            If a list of IDs is passed in,
            then the documents to find will be returned in the same order.

        Returns
        -------
            list[Document]: An empty list if no ducument is found.
        """
        
        documents = []
        for id in ids:
            document = self.find_document_by_id(id)
            documents.append(document)
        
        return documents
    
    @abstractmethod
    def update_documents(
            self, 
            documents: Iterable[Document],
            do_insert: bool = True
        ) -> list[DocumentID]:
        pass

    @abstractmethod
    def delete_documents(
            self, 
            documents: Iterable[Document]
        ) -> list[DocumentID]:
        pass
    
    def delete_document(
            self, 
            document: Document
        ) -> Optional[DocumentID]:
        
        deleted_document_ids = self.delete_documents([document])
        
        if len(deleted_document_ids) > 0:
            return deleted_document_ids[0]
        else:
            return None
    
    @abstractmethod
    def drop_collection(self, collection_name: str) -> None:
        pass



import unittest
from natia.kb.docdb import MongoDBClient

class DocumentDatabaseTester(unittest.TestCase):
    
    def test_():
        
        ...
        
if __name__ == '__main__':
    
    unittest.main()


from abc import ABC, abstractmethod
from typing import Optional, Iterable
from itertools import filterfalse
from ..document import Document, DocumentID

class BaseVectorDatabaseClient(ABC):
    
    @property
    @abstractmethod
    def collection_name(self) -> str:
        pass
    
    @staticmethod
    def filter_out_documents_without_ids(documents: Iterable[Document]) -> list[Document]:
        
        return list(filterfalse(
            lambda document: document.id is None,
            documents
        ))
        
    
    @abstractmethod
    def upsert_document(self, document: Document) -> Optional[DocumentID]:
        pass
    
    def upsert_documents(self, documents: Iterable[Document]) -> list[DocumentID]:
        
        document_ids = []
        for document in documents:
            document_id = self.upsert_document(document)
            if document_id is not None:
                document_ids.append(document_id)
        
        return document_ids
    
    @abstractmethod
    def delete_document_by_id(self, id: DocumentID) -> None:
        """Delete the vector embedding of the document specified by its ID.

        Paramerters
        -----------
            id (DocumentID): ID of the document to delete.
        """
        
        pass
    
    def delete_documents_by_ids(self, ids: Iterable[DocumentID]) -> None:
        """Delete all vector embeddings of the documents specified by their IDs.

        Paramerters
        -----------
            ids (Iterable[DocumentID]): IDs of the documents to delete.
        """
        
        for id in ids:
            self.delete_document_by_id(id)
        
    @abstractmethod
    def retrieve_similar_document_ids(
            self, 
            query: str, 
            n_similar_documents: int
        ) -> list[DocumentID]:
        pass
    
    @abstractmethod
    def drop_collection(self, collection_name: str) -> None:
        pass
        
        


import os
from uuid import UUID
from typing import Self, Optional, Iterable
from warnings import warn
from collections import namedtuple
import qdrant_client
from qdrant_client.models import (
    VectorParams, 
    Distance, 
    PointStruct,
    Batch,
    PointIdsList
)

from ...document import Document, DocumentID
from ..base import BaseVectorDatabaseClient
from ..doc2vec.doc2text import Document2TextTransformer
from ..doc2vec.base import BaseEmbeddingModel

# information of the collection
CollectionInfo = namedtuple(
    'CollectionInfo',
    (
        'name',
        'vec_dim',
        'metric'
    )
)

class QdrantClient(BaseVectorDatabaseClient, qdrant_client.QdrantClient):
    
    def __init__(self, *args, **kwargs):
        
        super().__init__(*args, **kwargs)
        
        # preserve super class's method create_collection
        self._create_collection = super().create_collection
        
        
        self._collection_name: Optional[str] = None

    @property
    def collection_name(self) -> Optional[str]:
        
        return self._collection_name
    
    @property
    def collections(self) -> list[str]:
        """Names of existing collections in the database.
        """
        
        # get a list of cellection descriptions
        collections = self.get_collections().collections
        
        # extract collection names
        collection_names = list(map(
            lambda collection: collection.name,
            collections
        ))
        
        return collection_names
     
    @classmethod
    def connect_to_host_port(cls, host: str, port: int) -> Self:
        """Connect to database via host and port.
        This requires that the docker image of Qdrant is running.

        Parameters
        ----------
            host (str): Host.
            port (int): Port.

        Returns
        -------
            Self: Client.
        """
     
        return cls(
            host=host,
            port=port
        )

    @classmethod
    def connect_to_dir(cls, dir: os.PathLike) -> Self:
        """Connect to local database specified by a directory path.
        In this case, the Qdrant database system is not used.
        Call this method only for testing APIs.

        Parameters
        ----------
            dir: Path to the directory for local storage.

        Returns
        -------
            Self: Client.
        """
        
        return cls(
            path=dir
        )
        
    def open_collection(
            self, 
            collection_name: str,
            document2text_transformer: Document2TextTransformer,
            embedding_model: BaseEmbeddingModel,
            do_create: bool = True,
            metric: Distance = Distance.COSINE
        ) -> Self:
        
        # handle the situation where the collection does not exist
        if not self.does_collection_exist(collection_name):
            
            # create a collection if it does not exist
            if do_create:
            
                self.create_collection(
                    collection_name=collection_name,
                    embedding_model=embedding_model,
                    metric=metric,
                    do_recreate=False,
                    do_warn=False
                )
            
            else:
                raise RuntimeError(f"collection '{collection_name}' does not exists")
        
        # set collection name
        self._collection_name = collection_name
        
        # set a transformer that transform a document to text
        # and the embedding model
        self._document2text_transformer = document2text_transformer
        self._embedding_model = embedding_model
        
        return self

    def does_collection_exist(self, collection_name: str) -> bool:
        """Check whether the given collection exists in the database.

        Parameters
        ----------
            collection (str): Collection name.

        Returns
        -------
            bool: True if the collection exists.
        """
        
        return collection_name in self.collections
    
    def create_collection(
            self, 
            collection_name: str,
            embedding_model: BaseEmbeddingModel,
            metric: Distance = Distance.COSINE,
            do_recreate: bool = False,
            do_warn: bool = False
        ) -> None:
        """Create a collection in Qdrant database.

        Parameters
        ----------
            name (str): Collection name.
            vec_dim (int): Dimension of the vector to store.
            metric (Distance, optional): Metric type, or in other words, 
            the distance function. 
            Defaults to Distance.COSINE.
            do_recreate (bool, optional): If set True, then the collection will be recreated/overwritten
            if it already exists in the database.
            Defaults to False.
            do_warn (bool, optional): If set True,
            then a warning message will be printed into terminal 
            if the collection already exists (and do_recreate is set False).
            Defaults to False.
        """
        
        # set the function to create a collection
        create_collection_func = qdrant_client.QdrantClient.create_collection
        
        if self.does_collection_exist(collection_name):
            
            # do nothing if the collection already exists
            if not do_recreate:
                
                # print warning message if required
                if do_warn:
                    warn(f"the collection '{collection_name}' already exists")
                    
                return
            
            # need to create the collection, 
            # and hence change the creation function
            create_collection_func = qdrant_client.QdrantClient.recreate_collection
        
        # use qdrant_client's API to create a collection
        create_collection_func(
            self,
            collection_name=collection_name,
            vectors_config=VectorParams(
                size=embedding_model.vec_dim, 
                distance=metric
            ),
        )
        
    def get_collection_info(self, collection_name: str) -> CollectionInfo | None:
        """Get the information of the collection.

        Parameters
        ----------
            collection_name (str): Collection name.

        Returns
        -------
            CollectionInfo | None: A CollectionInfo instance, which contains attributes:
            - name: str
            - vec_dim: int,
            - metric: Distance
        """
        
        # return None if the collection does not exist
        if not self.does_collection_exist(collection_name):
            return None
        
        # get collection configuration
        collection_config = self.get_collection(collection_name).config
        
        # get vector parameters
        vector_params = collection_config.params.vectors
        
        # dimension of the vector
        vec_dim = vector_params.size
        
        # metric used in the collection
        metric = vector_params.distance
        
        return CollectionInfo(name=collection_name, vec_dim=vec_dim, metric=metric)
    
    def upsert_document(self, document: Document) -> Optional[DocumentID]:
        
        if document.id is None:
            return None
        
        # get text content
        text = self._document2text_transformer.transform(document)
        
        # do nothing if the text cannot be extracted from the document
        if text is None:
            return None
        
        # insert into vector database
        self.upsert(
            collection_name=self.collection_name,
            points=[
                PointStruct(
                    id=str(document.id.to_uuid()),
                    vector=self._embedding_model.embed_text(text)
                )
            ]
        )
        
        return document.id
    
    def upsert_documents(
            self, 
            documents: Iterable[Document]
        ) -> list[DocumentID]:
        
        # filter out those documents without IDs
        documents = self.filter_out_documents_without_ids(documents)
        
        # get text contents
        texts = self._document2text_transformer.transform(documents)
        
        # IDs of the documents that have text contents
        document_ids = self._document2text_transformer.document_ids
        
        # stop here if there is no document to insert
        if len(document_ids) == 0:
            return []
        
        # convert to strings
        document_ids_as_strs = list(map(
            lambda document_id: str(document_id.to_uuid()), 
            document_ids
        ))
        
        # vector embeddings of text contents
        vectors = self._embedding_model.embed_texts(texts)
        
        # bulk insert into vector database
        self.upsert(
            collection_name=self.collection_name,
            points=Batch(
                ids=document_ids_as_strs,
                vectors=vectors
            )
        )
        
        return document_ids
        
    def delete_document_by_id(self, id: DocumentID) -> None:
        
        self.delete(
            collection_name=self.collection_name,
            points_selector=PointIdsList(
                points=[str(id.to_uuid())]
            )
        )
    
    def delete_documents_by_ids(self, ids: Iterable[DocumentID]) -> list[DocumentID]:
        
        # do nothing
        if len(ids) == 0:
            return []
        
        # convert IDs to a list of UUID strings
        document_ids = list(map(
            lambda document_id: str(document_id.to_uuid()),
            ids
        ))
        
        # find actual records to delete
        records = self.retrieve(
            collection_name=self.collection_name,
            ids=document_ids
        )
        
        # delete from database
        self.delete(
            collection_name=self.collection_name,
            points_selector=PointIdsList(
                points=document_ids
            )
        )
        
       # extract IDs
        deleted_document_ids = list(map(
            lambda record: DocumentID(UUID(record.id)),
            records
        ))
        
        return deleted_document_ids

    def retrieve_similar_document_ids(
            self, 
            query: str,
            n_similar_documents: int
        ) -> list[DocumentID]:
        
        # get the embedding vector of the query
        query_vector = self._embedding_model.embed_text(query)
        
        # get points with scores
        scored_points = self.search(
            collection_name=self.collection_name,
            query_vector=query_vector,
            limit=n_similar_documents,
            with_payload=False,
            with_vectors=False
        )
        
        # extract document IDs
        document_ids = list(map(
            lambda point: DocumentID(UUID(point.id)),
            scored_points
        ))
        
        return document_ids
        
    def drop_collection(self, collection_name: str) -> None:
        
        self.delete_collection(
            collection_name=collection_name
        )
    


from bson import ObjectId
from uuid import UUID
from typing import Type

class DocumentID:
    
    def __init__(self, value: ObjectId | UUID) -> None:
        
        self._value = value
        self._type = type(value)
        
    def __str__(self) -> str:
        return f'type: {self.type}, value: {self.value}'
    
    def __repr__(self) -> str:
        return str(self)
    
    @property
    def value(self) -> ObjectId | UUID:
        return self._value
    
    @property
    def type(self) -> Type:
        return self._type
    
    def to_objectid(self) -> ObjectId:
        
        if self.type is ObjectId:
            return self.value
        
        assert self.type is UUID
        return self.uuid_to_objectid(self.value)
    
    def to_uuid(self) -> UUID:
        
        if self.type is UUID:
            return self.value
        
        assert self.type is ObjectId
        return self.objectid_to_uuid(self.value)
    
    @staticmethod
    def objectid_to_uuid(objectid: ObjectId) -> UUID:
    
        objectid_12bytes = objectid.binary
        zero_4bytes = bytes(4)
        
        # 16 bytes required by UUID
        uuid_bytes = zero_4bytes + objectid_12bytes
        
        # create a UUID instance
        uuid = UUID(bytes=uuid_bytes)
        
        return uuid

    @staticmethod
    def uuid_to_objectid(uuid: UUID) -> ObjectId:
        
        # 16 bytes UUID
        uuid_bytes = uuid.bytes
        
        # remove the first 4 bytes
        objectid_12bytes = uuid_bytes.removeprefix(bytes(4))
        
        # create an ObjectId instance
        objectid = ObjectId(objectid_12bytes)
        
        return objectid
    